#include "lib.h"

TypeTree DB_TYPE;
NameTree DB_NAME;
Pool<Variable> DB_DATA;

extern "C" size_t type_outer(size_t type_id, size_t key)
{
    return DB_TYPE.outer(type_id, key);
}

extern "C" size_t type_inner(size_t type_id)
{
    return DB_TYPE.inner(type_id);
}

extern "C" size_t type_key(size_t type_id)
{
    return DB_TYPE.key(type_id);
}

extern "C" size_t type_size(size_t type_id)
{
    if(DB_TYPE.has(type_id)) {
        size_t type = DB_TYPE.key(type_id);

        switch(type) {
            case Type::Tag::Varying: return sizeof(Variable);
            case Type::Tag::Null:    return 0;
            case Type::Tag::Boolean: return sizeof(Type::Boolean);
            case Type::Tag::Natural: return sizeof(Type::Natural);
            case Type::Tag::Integer: return sizeof(Type::Integer);
            case Type::Tag::Block:   return DB_TYPE.key(DB_TYPE.inner(type_id));
            case Type::Tag::String:  return sizeof(Type::List);
            case Type::Tag::Array: {
                size_t length = DB_TYPE.inner(type_id);
                size_t inner = DB_TYPE.inner(length);
                return static_cast<size_t>(DB_TYPE.key(length)) * type_size(inner);
            };
            case Type::Tag::List:    return sizeof(Type::List);
            default: return 0;
        }
    }
    return 0;
}

void* allocate(size_t type_id, size_t count)
{
    void* mem = nullptr;
    size_t size = type_size(type_id) * count;
    if(size > 0) {
        mem = malloc(size);
        if(mem != nullptr) {
            memset(mem, 0, size);
        }
        return mem;
    }
    return nullptr;
}

extern "C" Variable acquire(size_t type_id)
{
    Variable addr {0};
    addr.address = allocate(type_id, 1);
    if(addr.address != nullptr) {
        addr.type = type_id;
    }
    return addr;
 }

void drop(Variable addr)
{
    if(addr.address != nullptr) {
        switch(addr.type) {
            case Type::Tag::Boolean:
            case Type::Tag::Natural:
            case Type::Tag::Integer:
            case Type::Tag::Block:
                break;

            case Type::Tag::Varying: {
                Variable& var = *reinterpret_cast<Variable*>(addr.address);
                if(var.address != nullptr) {
                    drop(var);
                }
                var.type = 0;
                var.address = nullptr;
            } break;

            case Type::Tag::List: {
                //Type::List& list = *reinterpret_cast<Type::List*>(addr.address);
                //if(list.data != nullptr) {
                //    for(size_t i = 0; i < list.length; ++i) {
                //        drop(list_cell(addr, i));
                //    }
                //    free(list.data);
                //}
            } break;
        }
        memset(addr.address, 0, type_size(addr.type));
    }
}

extern "C" void release(Variable addr)
{
    if(addr.address != nullptr) {
        drop(addr);
        free(addr.address);
    }
}

bool copy(Variable src, Variable dst)
{
    Variable source = src;
    Variable destination = dst;
    
    // dereference varying data
    if(type_key(src.type) == Type::Tag::Varying) {
        source = *reinterpret_cast<Variable*>(src.address);
    }

    // prepare destination for varying data
    if(type_key(dst.type) == Type::Tag::Varying) {
        auto& dest_ref = *reinterpret_cast<Variable*>(dst.address);
        
        // determine if memory can be reused, otherwise free and reallocate
        if(source.type != dest_ref.type) {
            if(dest_ref.address != nullptr) {
                free(dest_ref.address);
                dest_ref.type = Type::Tag::Null;
                dest_ref.address = nullptr;
            }
            dest_ref = acquire(source.type);
        }
    }

    // copy data into destination
    if(source.type == destination.type) {
        switch(type_key(destination.type)) {
            case Type::Tag::Null: { } break;

            case Type::Tag::Boolean:
            case Type::Tag::Natural:
            case Type::Tag::Integer:
            case Type::Tag::Block:
            {
                memcpy(destination.address, source.address, type_size(source.type));
            } break;
            
            /*case Type::Tag::List: {
                auto& src_list = *reinterpret_cast<Type::List*>(source.address);
                auto& dst_list = *reinterpret_cast<Type::List*>(destination.address);
                
                dst_list.capacity = src_list.capacity;
                dst_list.length = src_list.length;
                dst_list.data = allocate(type_inner(source.type), src_list.capacity);

                for(size_t i = 0; i < src_list.length; ++i) {
                    copy(list_at(source, i), list_at(destination, i));
                }
            } break;*/
        }

        return true;
    }

    return false;
}


// Varying //

extern "C" Variable varying_get(Variable addr)
{
    Variable result {0};
    //if(type_key(addr.type) == Type::Tag::Varying) {
        result = *reinterpret_cast<Variable*>(addr.address);
    //}
    return result;
}

extern "C" void varying_set(Variable addr, Variable source)
{
    //if(type_key(addr.type) == Type::Tag::Varying) {
        Variable& var = *reinterpret_cast<Variable*>(addr.address);
        
        if(var.address != nullptr) {
            drop(var);
        }
    
        if(var.type != source.type || var.address == nullptr) {
            if(var.address != nullptr) {
                free(var.address);
            }
            var.type = source.type;
            var.address = allocate(source.type, 1);
        }

        copy(source, var);
    //}
}

extern "C" void varying_clear(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::Varying) {
        Variable& var = *reinterpret_cast<Variable*>(addr.address);
        
        if(var.address != nullptr) {
            drop(var);
            free(var.address);
            var.type = 0;
            var.address = nullptr;
        }
    //}
}


// Boolean //

extern "C" void bool_set(Variable addr, Type::Boolean value)
{
    //if(type_key(addr.type) == Type::Tag::Boolean) {
        *(reinterpret_cast<Type::Boolean*>(addr.address)) = value;
    //}
}

extern "C" Type::Boolean bool_get(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::Boolean) {
        return *(reinterpret_cast<Type::Boolean*>(addr.address));
    //}
    //return false;
}


// Natural //

extern "C" void natural_set(Variable addr, Type::Natural value)
{
    //if(type_key(addr.type) == Type::Tag::Natural) {
        *(reinterpret_cast<Type::Natural*>(addr.address)) = value;
    //}
}

extern "C" Type::Natural natural_get(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::Natural) {
        return *(reinterpret_cast<Type::Natural*>(addr.address));
    //}
    //return false;
}


// Integer //

extern "C" void integer_set(Variable addr, Type::Integer value)
{
    //if(type_key(addr.type) == Type::Tag::Integer) {
        *(reinterpret_cast<Type::Integer*>(addr.address)) = value;
    //}
}

extern "C" Type::Integer integer_get(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::Integer) {
        return *(reinterpret_cast<Type::Integer*>(addr.address));
    //}
    //return false;
}


// Block //

extern "C" uint8_t block_get(Variable addr, size_t index)
{
    //if(type_key(addr.type) == Type::Tag::Block) {
        size_t length = type_key(type_inner(addr.type));
        if(index < length) {
            return reinterpret_cast<uint8_t*>(addr.address)[index];
        }
    //}
    return 0;
}

extern "C" void block_set(Variable addr, size_t index, uint8_t value)
{
    //if(type_key(addr.type) == Type::Tag::Block) {
        size_t length = type_key(type_inner(addr.type));
        if(index < length) {
            reinterpret_cast<uint8_t*>(addr.address)[index] = value;
        }
    //}
}


// String //


// Array //

extern "C" size_t array_length(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::Array) {
        return type_key(type_inner(addr.type));
    //}
    //return 0;
}

Variable array_cell(Variable addr, size_t index)
{
    Variable result {0};
    //if(type_key(addr.type) == Type::Tag::Array) {
        size_t length_n = type_inner(addr.type);
        size_t length = type_key(length_n);
        size_t type = type_inner(length_n);
        size_t offset = type_size(type);

        // validate for overflow
        if(addr.address != nullptr && offset > 0 && index < length) {
            result.type = type;
            result.address = reinterpret_cast<void*>(reinterpret_cast<size_t>(addr.address) + (offset * index));
        }
    //}
    return result;
}

extern "C" Variable array_get(Variable addr, size_t index)
{
    Variable result {0};
    //if(type_key(addr.type) == Type::Tag::Array) {
        Variable cell = array_cell(addr, index);
        if(cell.address != nullptr) {
            if(cell.type == Type::Tag::Varying) {
                result = varying_get(cell);
            }
            else {
                result = cell;
            }
        }
    //}
    return result;
}

extern "C" void array_set(Variable addr, size_t index, Variable source)
{
    //if(type_key(addr.type) == Type::Tag::Array) {
        Variable cell = array_cell(addr, index);
        if(cell.type == Type::Tag::Varying) {
            varying_set(cell, source);
        }
        else {
            copy(source, cell);
        }
    //}
}


// List //

extern "C" size_t list_capacity(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        return (*reinterpret_cast<Type::List*>(addr.address)).capacity;
    //}
    //return 0;
}

extern "C" size_t list_length(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        return (*reinterpret_cast<Type::List*>(addr.address)).length;
    //}
    //return 0;
}

Variable list_cell(Variable addr, size_t index)
{
    Variable result {0};
    //if(type_key(addr.type) == Type::Tag::List) {
        Type::List& list = *reinterpret_cast<Type::List*>(addr.address);
        size_t inner = type_inner(addr.type);
        size_t offset = type_size(inner);

        // validate for overflow
        if(list.data != nullptr && offset > 0 && index < list.length) {
            result.type = inner;
            result.address = reinterpret_cast<void*>(reinterpret_cast<size_t>(list.data) + (offset * index));
        }
    //}
    return result;
}

extern "C" Variable list_at(Variable addr, size_t index)
{
    Variable result {0};
    //if(type_key(addr.type) == Type::Tag::List) {
        Variable cell = list_cell(addr, index);
        if(cell.address != nullptr) {
            if(cell.type == Type::Tag::Varying) {
                result = *reinterpret_cast<Variable*>(cell.address);
            } else {
                result = cell;
            }
        }
    //}
    return result;
}

/*extern "C" Variable list_first(Variable addr)
{
    Variable result {0};
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        if(list.length > 0) {
            result = list_at(addr, 0);
        }
    }
    return result;
}*/

/*extern "C" Variable list_last(Variable addr)
{
    Variable result {0};
    if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        if(list.length > 0) {
            result = list_at(addr, list.length - 1);
        }
    }
    return result;
}*/

extern "C" void list_clear(Variable addr)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        size_t offset = type_size(type_inner(addr.type));
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        for(size_t i = 0; i < list.length; ++i) {
            drop(list_cell(addr, i));
        }
        list.length = 0;
        memset(list.data, 0, offset * list.capacity);
    //}
}

/*extern "C" void list_prepend(Variable addr, Variable source)
{
    if(type_key(addr.type) == Type::Tag::List) {
        size_t inner = type_inner(addr.type);
        //size_t offset = type_size(inner);

        // validate list can store value
        if(type_key(inner) == Type::Tag::Varying || source.type == inner) {
            auto& list = (*reinterpret_cast<Type::List*>(addr.address));

            if(list.length == list.capacity) {
                list_reserve(addr, list.capacity * 2);
            }

            // copy source to cell

            list.length++;
        }
    }
}*/

/*extern "C" void list_append(Variable addr, Variable source)
{
    if(type_key(addr.type) == Type::Tag::List) {
        Type::List& list = *reinterpret_cast<Type::List*>(addr.address);
        size_t inner = type_inner(addr.type);
        //size_t offset = type_size(inner);

        // validate list can store value
        if(type_key(inner) == Type::Tag::Varying || source.type == inner) {
            if(list.length == list.capacity) {
                list_reserve(addr, list.capacity * 2);
            }

            // copy source to cell

            list.length++;
        }
    }
}*/

extern "C" void list_insert(Variable addr, size_t index, Variable source)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        //auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    //}
}

extern "C" void list_update(Variable addr, size_t index, Variable source)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        //auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    //}
}

/*extern "C" void list_truncate(Variable addr, size_t maximum)
{
    if(type_key(addr.type) == Type::Tag::List) {
        //auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}*/

/*extern "C" void list_shift(Variable addr)
{
    if(type_key(addr.type) == Type::Tag::List) {
        //auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}*/

extern "C" void list_remove(Variable addr, size_t index, size_t maximum)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        //auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    //}
}

extern "C" void list_reserve(Variable addr, size_t capacity)
{
    //if(type_key(addr.type) == Type::Tag::List) {
        auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        size_t inner = type_inner(addr.type);
        size_t offset = type_size(inner);

        if(list.capacity < capacity) {
            void* new_data = allocate(inner, capacity);
            memcpy(new_data, list.data, offset * list.capacity);
        }
    //}
}

/*extern "C" void list_resize(Variable addr, size_t length)
{
    if(type_key(addr.type) == Type::Tag::List) {
        //auto& list = (*reinterpret_cast<Type::List*>(addr.address));
        
    }
}*/
