#ifndef H_NameTree
#define H_NameTree

#include <cstdint>
#include <vector>
#include <string>

class NameTree {
public:
    struct Node {
        std::string prefix;
        size_t parent;
        std::vector<size_t> children;
        
        Node(const std::string& prefix, size_t parent);
    };

    NameTree();
    ~NameTree();

    size_t lookup(const std::string& value);
    std::string get(size_t) const;

private:
    std::vector<Node> m_nodes;
};

NameTree::Node::Node(const std::string& prefix, size_t parent)
    :prefix(prefix), parent(parent) { }

NameTree::NameTree()
{
    m_nodes.push_back(Node("", 0));
}

NameTree::~NameTree()
{ }

size_t NameTree::lookup(const std::string& value)
{
    size_t current = 0;
    size_t index = 0;

    while(index < value.length()) { 
        Node& node = m_nodes[current];

        size_t find = 0;
        for(size_t ci = 0; ci < node.children.size(); ++ci) {
            size_t child = node.children[ci];
            if(value[index] == m_nodes[child].prefix[0]) {
                size_t prefix_index = 0;
                for(; prefix_index < m_nodes[child].prefix.length() && index < value.length(); prefix_index++ && index++) {
                    if(m_nodes[child].prefix[prefix_index] != value[index]) {
                        break;
                    }
                }

                if(prefix_index == m_nodes[child].prefix.length()) {
                    current = child;
                    break;
                }
                else {
                    find = child;

                    // create intermediate node
                    m_nodes.push_back(Node(m_nodes[child].prefix.substr(0, prefix_index), child));
                    m_nodes[current].children[ci] = m_nodes.size() - 1;
                    m_nodes[m_nodes.size() - 1].children.push_back(child);
                    
                    // update child node
                    m_nodes[child].parent = m_nodes.size() - 1;

                    // create branching node if value is not prefix
                    if(index != value.length()) {
                        m_nodes.push_back(Node(value.substr(index), child));
                        find = m_nodes.size() - 1;
                    }
                    
                    return find;
                }
            }
        }

        if(find == 0) {
            // add suffix to new node pointing to value
            m_nodes.push_back(Node(value.substr(index), current));
            find = m_nodes.size() - 1;
            node.children.push_back(find);
            return find;
        }
    }

    return current;
}

std::string NameTree::get(size_t id) const
{
    size_t length = 0;
    std::string out;

    if(id < m_nodes.size()) {
        for(size_t current = id; current != 0;) {
            length += m_nodes[current].prefix.length();
            current = m_nodes[current].parent;
        }

        out = std::string(length, '\0');
        size_t index = length - 1;

        for(size_t current = id; current != 0;) {
            size_t prefix_length = m_nodes[current].prefix.length() - 1;
            for(size_t i = 0; i < m_nodes[current].prefix.length(); ++i) {
                out[index--] = m_nodes[current].prefix[prefix_length - i];
            }
            current = m_nodes[current].parent;
        }
    }

    return out;
}

#endif
